#!/usr/bin/perl
###############################################################################
#
# File:         reg_reset.pl
# Description:  Extract register reset and write mask info from documentation
# Author:       Matthew Lovell
# Created:      Fri Jul 16 12:44:21 1999
# Modified:     Thu Sep  6 13:07:33 2001
# Language:     CPerl
# Package:      register
# License:      LPPL
#
# (C) Copyright 2001, Matthew Lovell  (lovell@fc.hp.com)
#
###############################################################################
#
#
# This script recursively reads .tex files (recursing upon
# encountering an \include and \input) and finds any register macros.
# It extracts the field information from each macro and adds it to a
# growing data structure.
#
# The resulting data can be used to create register information
# suitable for metachecking purposes.
#
# The parsing of the .tex files is not necessary the most rigorous,
# just sufficient for the job! :)
#
#
# USAGE
#
#  reg_reset.pl <top-level TeX file>
#
#    STDERR will contain parsing and debug information
#    STDOUT should be an e file suitable for use.
#   
#
###############################################################################

select STDERR;
use FileHandle;

# Structure to hold register definitions from documentation
@REGDEF = ();

# Structure to recognize register offset collisions
@OFFSETS = ();

###############################################################################
#                               M A I N                                       #
###############################################################################
$createfieldlist=0;
if ($createfieldlist) {
  open (WFIELDLIST,">wfieldlist");
  open (NWFIELDLIST,">nwfieldlist");
  print WFIELDLIST "Writeable Fields of Registers.\n";
  print NWFIELDLIST "Non-writeable Fields of Registers.\n";
};

my @files = get_tex_list($ARGV[0]);

for (@files)  { get_regs($_) }

print STDOUT "Register definitions\n";
print STDOUT "Created " . localtime() . " by $0 for $ENV{LOGNAME} \n\n";

my $header = sprintf "def_mreg %-18s %-8s %-20s %-20s %-20s",
  "NAME","OFFSET","RESET","RESET_MASK","WRITE_MASK";
print STDOUT "$header\n";
print STDOUT "-" x 98, "\n";  

print STDOUT "<'\n";
for (sort {hex($a->{offset}) <=> hex($b->{offset})} @REGDEF) { process_reg($_) }
print STDOUT "'>\n";

###############################################################################
#                         S U B R O U T I N E S                               #
###############################################################################


# Given a register definition, produce an appropriate line of output.
# Each line should have the register name, offset, reset value,
# reset mask, and write mask.
#
# Arguments:  reference to hash with the following fields
#
#               long_name        Full register name
#               short_name       Register abbreviation
#               offset           Offset
#               testable         Indicates register is 'normal'
#               fields           reference to array of hashes, each contains
#
#                                  name        Field name
#                                  len         Field length
#                                  start       Starting bit
#                                  reset       Reset value (may not be valid)
#                                  reset_mask  Indicates field is valid after reset
#                                  write_mask  Indicates field is updated via write
#               
#
sub process_reg {
  my ($reg) = @_;

  my $reset_value = "0" x 64;
  my $reset_mask  = "1" x 64;
  my $write_mask  = "1" x 64;

  print "Working on $reg->{short_name} ($reg->{offset})  ";

  if ($reg->{testable} == 0) {
    $reset_mask = "0" x 64;
    $write_mask = "0" x 64;
    print "DONT_TEST";
  }

  print "\n";

  my $field;
  foreach $field (@{$reg->{fields}}) {
    my $value = 0;
    my $valid = 1;

    # remove spaces
    $field->{reset} =~ s/\s*//g;

    # binary    
    if ($field->{reset} =~ /^[01]+$/ ) {
      $value = substr ("0" x 32 . $field->{reset}, -32);
    } 
    # hexadecimal
    elsif ($field->{reset} =~ /^0?x[0-9A-F]+$/i ) {
      $value = hex2bin($field->{reset});
    } 
    # no reset value
    else {
      print "  No reset value for '$field->{name}'\n";
      $field->{reset_mask} = 0;
      $valid = 0;
    }

    if ($valid) {
      # value is currently 32 characters long.  Isolate the desired lower bits
      $value = substr $value, - $field->{len};
      print "  '$field->{name}' resets to $value\n";
      # insert into proper location in bit string
      substr($reset_value,
	     -($field->{start}+$field->{len}), 
	     $field->{len}) = $value;
    }

    # treat reset and write masks similarly
    substr($reset_mask, 
	   -($field->{start}+$field->{len}), 
	   $field->{len}) = "$field->{reset_mask}" x $field->{len};
    
    substr($write_mask,
	   -($field->{start}+$field->{len}), 
	   $field->{len}) = "$field->{write_mask}" x $field->{len};
    
  }
  
  print "\n";
  print "  RESET: 64'b$reset_value\n";
  print "  RMASK: 64'b$reset_mask\n";
  print "  WMASK: 64'b$write_mask\n\n";

  print "  RESET: 64'h", bin2hex($reset_value), "\n";
  print "  RMASK: 64'h", bin2hex($reset_mask), "\n";
  print "  WMASK: 64'h", bin2hex($write_mask), "\n\n";


  # Only take those registers with valid offsets, recognizing ranges
  if ($reg->{offset} =~ /^(0?x[0-9A-F]+)(?:\s*-+\s*(0?x[0-9A-F]+))?$/i) {

    my $starting_offset = hex($1);
    my $ending_offset   = $starting_offset;

    $ending_offset = hex($2) if (defined $2);
    
    my $index;
    my $iteration = 0;
    my $short_name = "";
    for ($index = $starting_offset; $index <= $ending_offset; $index += 0x0008) {
      if (defined $OFFSETS{$index}) {
	printf STDERR "## ERROR  Register collision detected at offset %04x\n", $index;
	printf STDOUT "-- ERROR: Register collision detected at offset %04x\n", $index;
      } else {
	$OFFSETS{$index}++;
      }
      my $short_name = $reg->{short_name};
      $short_name .= $iteration++ unless ($starting_offset == $ending_offset);
      my $output = 
	sprintf "def_mreg %-18s 16'h%04x %20s %20s %20s;",
	$short_name,
	$index, 
	"64'h" . bin2hex($reset_value),
	"64'h" . bin2hex($reset_mask),
	"64'h" . bin2hex($write_mask);
      
      print STDOUT "$output\n";
    }
  }


  print "\n";
  
} # process_reg


# Convert an ASCII string representing a binary number to a true decimal number
sub bin2dec {
  return unpack("N", pack("B32", substr("0" x 32 . shift, -32)));
}

# Convert a true number to a binary ASCII string
sub dec2bin {
  my $str = unpack("B32", pack("N", shift));
  return $str;
}

# Covert an ASCII string representing a hex number to a binary ASCII string;
sub hex2bin {
  return unpack("B32", pack("N", hex(shift)));
}

# Convert an ASCII string representing a binary number to a hex ASCII string;
sub bin2hex {

    my $Bin = shift(@_);
    my $Hex;

    # Clean up string and look for errors
    chomp $Bin;

    if ($Bin =~ m/[^10]/) {
        print STDERR "ERROR: $Bin is not binary\n";
        return ("");
    }
    return scalar reverse unpack "h*",(pack "b*", scalar reverse $Bin);
}



# Read a .tex file line by line, looking for register declarations
# using the register.sty macro.  This subroutine only examines
# a single file.
# 
# Arguments: filename
#
sub get_regs {
  my ($filename) = @_;

  my $file = new FileHandle "$filename";

  while (<$file>) {
    # find register definition
    if ( m/\\begin\{register\}\{[\w\!]+\}\{(.*)\}\{(.*)\}%(.*)$/ ) {
      my $long_name  = $1;
      my $offset     = $2;
      my $testable   = 1;
      my $short_name = "";

      my $remainder  = $3;

      if ($remainder ne "") {
	if ($remainder =~ /don\'?t(_| )test/i) { $testable = 0 }
	if ($remainder =~ /NAME=(\w+)/) { $short_name = $1 }
      }

      if ($short_name eq "") {
	$short_name = $long_name;
	$short_name =~ tr/ /_/;
	$short_name =~ tr/a-z/A-Z/;
      }

      my $width = 0;
      my $reg = {
		 long_name  => $long_name,
		 offset     => $offset,
		 testable   => $testable,
		 short_name => $short_name,
		 fields     => []				
		};

      print STDERR "-" x 70, "\n";
      print STDERR
	"Parsed register '$long_name'\n",
	"  offset     : $offset\n",
	"  testable   : $testable\n",
	"  short name : $short_name\n\n";

      # extract fields
      # - read in whole rest of the reg macro
      @wholereg = ();
      while (<$file>) {
	push(@wholereg, $_);
	last if (/\\end\{register\}/);
      }

      # combine lines if nested {}'s do not match
      for (my $i = 0; $i <= $#wholereg; $i++) {
	my $open  = grep(/\{/, split(//, $wholereg[$i]));
	my $close = grep(/\}/, split(//, $wholereg[$i]));

	if ($open != $close) {
	  my ($next) = splice(@wholereg, $i+1, 1);
	  $wholereg[$i] .= $next;
	  redo;
	}
      }

      foreach (@wholereg) {
	if ( m/\\regfield\{(.*)\}\s*\{(.*)\}\s*\{(.*)\}\s*\{(.*)\}\s*%(.*)$/s ) {
	  my $remain = $5;
	  my $field_name  = $1;
	  my $field_len   = $2;
	  my $field_start = $3;  
	  my $reset       = $4;	  $reset =~ s/^\{//;  $reset =~ s/\}$//;  $reset =~ s/_//g;
	  my $reset_mask  = $testable;	  # indicates field is deterministic after reset
	  my $write_mask  = $testable;    # indicates field is modified by regbus writes


	  if ($field_name =~ m/\{([\w ])\}$/) {
	    $field_name = $1;
	  }
	  $field_name =~ s/\\_/_/g;


	  if ($remain =~ /read(?:_| )only/i) { $write_mask = 0; }
	  if ($remain =~ /status/i)          { $reset_mask = 0; $write_mask = 0; }
	  if ($field_name =~ /reserved/i)    { $write_mask = 0; }


	  if ($field_len > 32) {
	    die "ERROR! Field of greater than 32 bits in $short_name!\n";	    
	  }

	  print STDERR 
	    "  Parsed field '$field_name'\n",
	    "    length     : $field_len\n",
	    "    start      : $field_start\n",
	    "    reset      : $reset\n",
	    "    reset_mask : $reset_mask\n",
	    "    write_mask : $write_mask\n\n";

	  if($createfieldlist) { #added to get info for test writing
	    my $field_end=$field_start+$field_len-1;
	    if ($write_mask) {
	      print WFIELDLIST "$short_name\[$field_end\:$field_start\] $field_name\n"; 
	    } else {
	      print NWFIELDLIST "$short_name\[$field_end\:$field_start\] $field_name\n"; 
	    };
	  };
	  push(@{$reg->{fields}},
	       {
		name  => $field_name,
		len   => $field_len,
		start => $field_start,
		reset => $reset,
		reset_mask => $reset_mask,
		write_mask => $write_mask
	       });
	}
      } # end of register macro found
      
      push(@REGDEF, $reg);
    }
  }

} # get_regs


# Read a .tex file line by line, finding all \include or \input
# commands.  Recurse upon encountering one of those, thus making
# a list of the entire TeX file tree.
#
# Arguments: filename
#
sub get_tex_list {
  my ($filename) = @_;

  # Add .tex if not present
  if ($filename !~ /\.\w+$/) {
    $filename .= '.tex';
  }

  # initialize file list
  my @file_list = ($filename);

  my $file = new FileHandle "$filename";  
  while (<$file>) {

    my @list = m/\\(?:include|input)\{([\w\/.]+)\}/g;

    my $subfile;
    for $subfile (@list) { 
      push(@file_list, get_tex_list($subfile));
    }
  }

  return @file_list;
} # get_tex_list






