#!/usr/bin/perl
###############################################################################
#
# File:         reg_table.pl
# Description:  Produce a register list in offset order
# Author:       Matthew Lovell
# Created:      Wed May 12 13:35:18 1999
# Modified:     Thu Sep  6 13:07:28 2001
# Language:     CPerl
# Package:      register
# License:      LPPL
#
# (C) Copyright 2001, Matthew Lovell  (lovell@fc.hp.com)
#
###############################################################################
#
#
# This script is meant to be passed the .rdf (List of Registers) file
# from a LaTeX document using the register package, i.e.
#
# USAGE
#
#    reg_table.pl <document>.rdf > ordered_regs.data
#
# On STDOUT, it produces LaTeX code suitable for the *rows* of a table
# listing all encountered registers in offset order.  For space that
# should be reserved, it colors the corresponding row gray and marks
# it as such.
#
###############################################################################


#
# Name for unused register space.  Typical strings would be
# 'Unimplemented' or 'Reserved'
#
$UNUSED = 'Reserved';

# Hash storing register space allocation.  Mark reserved or unimplemented
# regions here.  The label for such areas should specific be written as
# 'Reserved' or 'Unimplemented'.  Use only capital letters for any hex
# digits.  Indicate ranges using ' - ', including the spaces.
#
# All addresses used should be the _starting_ address registers, i.e.
# 8-byte aligned.  Only use ranges when indicating more than one
# register
#
%REG = 
 (
  '0x0010' => { offset => '0x0010 - 0x0028', label => 'Reserved' },

  # Old register
  '0x0050' => { offset => '0x0050',          label => 'Reserved' },

  # Suspend register
  '0x0078' => { offset => '0x0078',          label => 'Reserved' },
  
  # Arbitration Mode
  '0x0088' => { offset => '0x0088',          label => 'Reserved' },
  '0x0110' => { offset => '0x0110 - 0x01F8', label => 'Reserved' },

  # Unimplemented space
  '0x03E0' => { offset => '0x03E0 - 0x03E8', label => 'Reserved' },

  # Old registers
  '0x0240' => { offset => '0x0240 - 0x0248', label => 'Reserved' },
  '0x0260' => { offset => '0x0260 - 0x0268', label => 'Reserved' },

  # PCI Drive strength
  '0x0608' => { offset => '0x0608',          label => 'Reserved' }
 );


# Define address of the last byte of the last possible register in the device
$UPPER_LIMIT = 0x0FFF;


###############################################################################
#                               M A I N                                       #
###############################################################################

# Read in file
while (<>) {

  # skip over spaces
  next if (m/^\\addvspace/);

  my $key;
  my ($name, $offset) = 
    m/
      \\textsc
        \s+\{
           ([^\(]+)           # register name
        \}\s
           \(\{(.+)\}\)       # register offset
     /x;
  
  # recognize ranges
  if ($offset =~ m/(0x[0-9A-F]+)\s*-+\s*(0x[0-9A-F]+)/i) {
    $key = $1;
    $offset = "$1 - $2";
  } else {
    $key = $offset;
  }

  # produce a decent label
  ($label = $name) =~ s/\{|\}//g;
  $offset =~ tr/a-f/A-F/;

  # detect collisions


  if (defined $REG{$key}) {
    $REG{$key}{label} = "COLLISION!";
    print STDERR "#ERROR  Collision detected at offset ", $key, " !!!\n";
  } else {
    $REG{$key} = {
		  label  => $label,
		  name   => $name,
		  offset => $offset
		 };
  }

}

# Output in useful form
produce_table();

###############################################################################
#                         S U B R O U T I N E S                               #
###############################################################################

# Produce rows suitable for inclusion into a table or longtable
sub produce_table {
  my $index;
  my $first_time = 1;
  my $in_unimplemented = 0;
  my $unimp_start = "";
  my $unimp_end = "";
  print " \n";

  # Go through register space
  for ($index = 0; $index < $UPPER_LIMIT; $index += 0x0008) {
    my $offset = sprintf "0x%04X", $index;
    if (defined $REG{$offset}) {

      # Finish off unimplemented range
      if ($in_unimplemented) {
	$unimp_end = sprintf "0x%04X", $index - 0x0008;
	if ($unimp_end eq $unimp_start) {
	  $REG{$unimp_start} = {
				label  => $UNUSED,
				offset => "$unimp_start",
				visited => 1
			       };
	} else {
	  $REG{$unimp_start} = {
				label  => $UNUSED,
				offset => "$unimp_start - $unimp_end",
				visited => 1
			       };
	}
	print " \n" unless ($first_time);
	if ($first_time) { $first_time = 0; }
	produce_row($unimp_start);
	$in_unimplemented = 0;
      }

      printf STDERR "0x%04X %s\n", $index, $REG{$offset}{label};
      print " \n" unless ($first_time);
      if ($first_time) { $first_time = 0; }
      $REG{$offset}{visited}++;
      produce_row($offset);
      
      # Recognize ranges
      if ($REG{$offset}{offset} =~ m/0x[0-9A-F]+\s*-+\s*(0x[0-9A-F]+)/i) {
	# jump to end of range
	$index = hex($1);
	print STDERR "Encountered range $REG{$offset}{offset}, ";
	printf STDERR "skipping to 0x%04X\n", $index;
      }

    } else {
      printf STDERR "0x%04X seems empty...\n",$index;
      if (!$in_unimplemented) {
	# Start unimplemented range
	$unimp_start = sprintf "0x%04X", $index;
	$in_unimplemented = 1;
      }
    }
  }

  # Take care of end of list
  if ($in_unimplemented) {
	$unimp_end = sprintf "0x%04X", $index - 0x0008;
	if ($unimp_end eq $unimp_start) {
	  $REG{$unimp_start} = {
				label  => $UNUSED,
				offset => "$unimp_start",
				visited => 1
			       };
	} else {
	  $REG{$unimp_start} = {
				label  => $UNUSED,
				offset => "$unimp_start - $unimp_end",
				visited => 1
			       };
	}
	print " \n" unless ($first_time);
	if ($first_time) { $first_time = 0; }
	produce_row($unimp_start);
	$in_unimplemented = 0;
  }


  # Find any register we didn't hit
  my $key; 
  foreach $key (keys %REG) {
    unless ($REG{$key}{visited}) {
      print " \n";
      produce_row($key);
    }
  }  
} # produce table


# Produce an individual row in the table
sub produce_row {
  my ($key) = @_;

  if ($REG{$key}{label} =~ /Reserved|Unimplemented/) {
    print "\\rowcolor[gray]{0.9}\n";
    print "\\texttt{$REG{$key}{offset}} & {\\it $REG{$key}{label}} & \\\\ \n";
  } else {
    print "\\texttt{$REG{$key}{offset}} & $REG{$key}{label} & \\pageref{$REG{$key}{name}} \\\\ \n";
  }
} # produce_row
